@@ -21,7 +21,7 @@ import * as vscode from 'vscode';
 import { CancellationError, CancellationToken, DocumentSymbol, SymbolKind } from 'vscode';
 import { LanguageClient } from 'vscode-languageclient/node';
 import winston from 'winston';
-// eslint-disable-next-line @typescript-eslint/no-unused-vars
+
 import { existsSync } from 'fs';
 import { EXTENSION_NAME, adaExtState, logger, mainOutputChannel } from './extension';
 
@@ -71,14 +71,14 @@ export function substituteVariables(str: string, recursive = false): string {
     str = str.replace(/\${relativeFile}/g, relativeFilePath);
     str = str.replace(
         /\${relativeFileDirname}/g,
-        relativeFilePath.substr(0, relativeFilePath.lastIndexOf(path.sep))
+        relativeFilePath.substr(0, relativeFilePath.lastIndexOf(path.sep)),
     );
     str = str.replace(/\${fileBasename}/g, parsedPath.base);
     str = str.replace(/\${fileBasenameNoExtension}/g, parsedPath.name);
     str = str.replace(/\${fileExtname}/g, parsedPath.ext);
     str = str.replace(
         /\${fileDirname}/g,
-        parsedPath.dir.substr(parsedPath.dir.lastIndexOf(path.sep) + 1)
+        parsedPath.dir.substr(parsedPath.dir.lastIndexOf(path.sep) + 1),
     );
     str = str.replace(/\${cwd}/g, parsedPath.dir);
     str = str.replace(/\${pathSeparator}/g, path.sep);
@@ -88,18 +88,16 @@ export function substituteVariables(str: string, recursive = false): string {
         str = str.replace(
             /\${selectedText}/g,
             activeEditor.document.getText(
-                new vscode.Range(activeEditor.selection.start, activeEditor.selection.end)
-            )
+                new vscode.Range(activeEditor.selection.start, activeEditor.selection.end),
+            ),
         );
     }
 
     str = str.replace(/\${env:(.*?)}/g, function (variable) {
-        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
         return process.env[variable.match(/\${env:(.*?)}/)![1]] || '';
     });
 
     str = str.replace(/\${config:(.*?)}/g, function (variable) {
-        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
         return vscode.workspace.getConfiguration().get(variable.match(/\${config:(.*?)}/)![1], '');
     });
 
@@ -107,7 +105,7 @@ export function substituteVariables(str: string, recursive = false): string {
         recursive &&
         str.match(
             // eslint-disable-next-line max-len
-            /\${(workspaceFolder|workspaceFolderBasename|fileWorkspaceFolder|relativeFile|fileBasename|fileBasenameNoExtension|fileExtname|fileDirname|cwd|pathSeparator|lineNumber|selectedText|env:(.*?)|config:(.*?))}/
+            /\${(workspaceFolder|workspaceFolderBasename|fileWorkspaceFolder|relativeFile|fileBasename|fileBasenameNoExtension|fileExtname|fileDirname|cwd|pathSeparator|lineNumber|selectedText|env:(.*?)|config:(.*?))}/,
         )
     ) {
         str = substituteVariables(str, recursive);
@@ -122,6 +120,23 @@ export const TERMINAL_ENV_SETTING_NAME =
     'terminal.integrated.env.' +
     (platform() == 'darwin' ? 'osx' : platform() == 'win32' ? 'windows' : 'linux');
 
+/**
+ * Check if VS Code is running in a remote environment.
+ * @returns True if running remotely, false otherwise.
+ */
+export function isRunningOnRemote(): boolean {
+    return vscode.env.remoteName !== undefined;
+}
+
+/**
+ * Check if a VS Code extension is installed.
+ * @param extensionId - The ID of the extension to check.
+ * @returns True if the extension is installed, false otherwise.
+ */
+export function isExtensionInstalled(extensionId: string): boolean {
+    return vscode.extensions.getExtension(extensionId) !== undefined;
+}
+
 /**
  *
  * @returns the value of the applicable `terminal.integrated.env.*` setting,
@@ -140,7 +155,7 @@ export function getTerminalEnv(): { [name: string]: string | null } {
  * @returns the value of the applicable `terminal.integrated.env.*` setting,
  * after evaluation of macros such as `${env:...}`.
  */
-export function getEvaluatedTerminalEnv() {
+export function getEvaluatedTerminalEnv(): { [name: string]: string | null } {
     const custom_env = getTerminalEnv();
 
     if (custom_env) {
@@ -154,7 +169,7 @@ export function getEvaluatedTerminalEnv() {
                 // in the JSON settings configuration (e.g: "PATH": "${workspaceFolder}/obj")
                 custom_env[var_name] =
                     custom_env[var_name]?.replace(/(\$\{.*\})/, (substring) =>
-                        substituteVariables(substring, false)
+                        substituteVariables(substring, false),
                     ) ?? null;
             }
         }
@@ -181,7 +196,7 @@ export function getArgValue(a: string | vscode.ShellQuotedString): string {
  */
 export function setTerminalEnvironment(
     targetEnv: NodeJS.ProcessEnv,
-    custom_env?: { [name: string]: string | null }
+    custom_env?: { [name: string]: string | null },
 ) {
     if (custom_env == undefined) {
         // Retrieve the user's custom environment variables if specified in their
@@ -243,7 +258,7 @@ export function assertSupportedEnvironments(mainChannel: winston.Logger) {
     }
 
     logger.debug(
-        `Asserted compatibility with runtime environment: ${process.arch}, ${process.platform}`
+        `Asserted compatibility with runtime environment: ${process.arch}, ${process.platform}`,
     );
 }
 
@@ -321,7 +336,7 @@ export async function getAdaMains(): Promise<AdaMain[]> {
         execs.length == mains.length,
         `The ALS returned mains.length = ${mains.length} and ` +
             `execs.length = ${execs.length}` +
-            `when they should be equal`
+            `when they should be equal`,
     );
 
     const result: AdaMain[] = [];
@@ -395,12 +410,12 @@ export async function findAdaMain(mainPath: string): Promise<AdaMain | undefined
         const adaMain = projectMains.find(
             (val) =>
                 val.mainRelPath().toLowerCase() == lc_main ||
-                val.mainFullPath.toLowerCase() == lc_main
+                val.mainFullPath.toLowerCase() == lc_main,
         );
         return adaMain;
     } else {
         const adaMain = projectMains.find(
-            (val) => val.mainRelPath() == mainPath || val.mainFullPath == mainPath
+            (val) => val.mainRelPath() == mainPath || val.mainFullPath == mainPath,
         );
         return adaMain;
     }
@@ -425,7 +440,7 @@ export function getSymbols(
     rootSymbols: DocumentSymbol[],
     symbolKinds: SymbolKind[],
     recurseInto: SymbolKind[] = [SymbolKind.Module, SymbolKind.Package, SymbolKind.Function],
-    token?: CancellationToken
+    token?: CancellationToken,
 ): DocumentSymbol[] {
     const reduce = (acc: DocumentSymbol[], cur: DocumentSymbol) => {
         if (token?.isCancellationRequested) {
@@ -494,7 +509,7 @@ export function which(execName: string) {
                * On Windows use a default list of extensions in case PATHEXT is
                * not set.
                */
-              env.PATHEXT?.split(path.delimiter) ?? ['.exe', '.cmd', '.bat']
+              (env.PATHEXT?.split(path.delimiter) ?? ['.exe', '.cmd', '.bat'])
             : [''];
 
     const exePath: string | undefined = paths
@@ -513,7 +528,7 @@ export function which(execName: string) {
  */
 export async function showErrorMessageWithOpenLogButton(
     msg: string,
-    options?: vscode.MessageOptions
+    options?: vscode.MessageOptions,
 ) {
     const answer = await vscode.window.showErrorMessage<vscode.MessageItem>(msg, options ?? {}, {
         title: `Open ${EXTENSION_NAME} Extension Log`,
@@ -522,3 +537,193 @@ export async function showErrorMessageWithOpenLogButton(
         mainOutputChannel.show();
     }
 }
+
+/**
+ *
+ * @param data - input data to process
+ * @param threads - number of async functions processing the data. Must be strictly positive 0.
+ * @param op - the function to process the data
+ * @returns the processed
+ */
+export async function parallelize<T, Result>(
+    data: T[],
+    threads: number,
+    op: (arg: T) => Result | Promise<Result>,
+    token?: CancellationToken,
+): Promise<Result[]> {
+    assert(threads > 0);
+    const len = data.length;
+    let start = 0;
+
+    /**
+     * Initialize an array with the expected length.
+     */
+    const result = new Array<Result>(len);
+    while (start < len) {
+        if (token?.isCancellationRequested) {
+            throw new CancellationError();
+        }
+
+        /**
+         * Take a chunk corresponding to the number of threads.
+         */
+        const chunk = data.slice(start, start + threads);
+
+        /**
+         * Process the chunk as async promises which may be evaluated concurrently.
+         */
+        const chunkRes = await Promise.all(chunk.map(op));
+
+        /**
+         * Since the array is initialized with a length, its content is already
+         * allocated with empty values. So we need a splice operation where we
+         * delete the existing empty values and replace them with the results.
+         */
+        result.splice(start, chunkRes.length, ...chunkRes);
+
+        start += threads;
+    }
+
+    return result;
+}
+
+/**
+ * When using {@link vscode.window.withProgress}, reporting progress too often
+ * can cause the UI to freeze. This function helps stagger progress reports
+ * such that progress is reported only when a minimum increment is reached,
+ * e.g. every 1%.
+ *
+ * @example Here's how the function should be used:
+ * ```ts
+ * await vscode.window.withProgress(...,
+ *     async (progress, token) => {
+ *         let done: number = 0;
+ *         let lastProgress = 0;
+ *         const total = ...;
+ *
+ *         for (...) {
+ *             lastProgress = staggerProgress(
+ *                 done,
+ *                 total,
+ *                 lastProgress,
+ *                 (increment) => {
+ *                     progress.report({
+ *                         message: `${done} / ${total}`,
+ *                         increment: increment,
+ *                     });
+ *                 },
+ *             );
+ *             ++done;
+ *         }
+ *     }
+ * })
+ * ```
+ *
+ * @param done - the number of items processed since the beginning of the work
+ * @param total - the total number of items to be processed
+ * @param lastProgress - the last 'done' value at which progress was reported
+ * @param progress - a callback that is called when the minimum increment is reached
+ * @param minimumIncrement - the minimum increment to wait for before calling
+ * the 'progress' callback.
+ * @returns `lastProgress` if no progress was reported, and returns `done` when
+ * progress was reported.
+ */
+export function staggerProgress(
+    done: number,
+    total: number,
+    lastProgress: number,
+    progress: (increment: number) => void,
+    minimumIncrement: number = 1,
+): number {
+    const sinceLastReport = done - lastProgress;
+    const increment = Math.floor((sinceLastReport * 100) / total);
+    if (increment >= minimumIncrement) {
+        progress(increment);
+        return done;
+    } else {
+        return lastProgress;
+    }
+}
+
+/**
+ *
+ * @param arg - a string
+ * @returns a modified string suitable for use as a filename
+ */
+export function slugify(arg: string): string {
+    return arg.replace(/[<>:"/\\|?*]/g, '_');
+}
+/**
+ * @returns the length of the common suffix between the given strings. If there
+ * is no common suffix, the result is `0`.
+ */
+export function getLengthCommonSuffix(str1: string, str2: string): number {
+    /**
+     * Compare characters from the end of each string and iterating backwards.
+     */
+    let len = 0;
+    while (
+        len < str1.length &&
+        len < str2.length &&
+        str1[str1.length - 1 - len] == str2[str2.length - 1 - len]
+    ) {
+        len++;
+    }
+
+    return len;
+}
+
+/**
+ *
+ * @param p - a path
+ * @returns a POSIX version of the same path obtained by replacing occurences
+ * of `\` with `/`. If the input path was a Windows absolute path, a `/` is
+ * prepended to the output to make it also an absolute path.
+ */
+export function toPosix(p: string) {
+    let posixPath = p.replace(RegExp(`\\${path.win32.sep}`, 'g'), path.posix.sep);
+
+    /**
+     * If it was an absolute path from Windows, we have to
+     * manually make it a POSIX absolute path.
+     */
+    if (path.win32.isAbsolute(p) && !path.posix.isAbsolute(posixPath)) {
+        posixPath = `/${posixPath}`;
+    }
+
+    return posixPath;
+}
+
+/**
+ *
+ * @param path1 - a path
+ * @param path2 - another path
+ * @returns if the given paths end with a common suffix, the remaining prefixes
+ * are returned as a tuple by this function. Otherwise, undefined prefixes are
+ * returned.
+ */
+export function getMatchingPrefixes(
+    path1: string,
+    path2: string,
+): [string | undefined, string | undefined] {
+    const lenCommonSuffix = getLengthCommonSuffix(path1, path2);
+
+    let prefix1;
+    let prefix2;
+
+    if (lenCommonSuffix > 0) {
+        prefix1 = path1.slice(0, path1.length - lenCommonSuffix);
+        prefix2 = path2.slice(0, path2.length - lenCommonSuffix);
+    } else {
+        // Could not find a common suffix
+        prefix1 = prefix2 = undefined;
+    }
+
+    return [prefix1, prefix2];
+}
+
+/**
+ * This constant indicates whether the code is running in a testing
+ * environment.
+ */
+export const inTesting = process.env['ALS_VSCODE_TEST_ENV'] === '1';
